import threading
import asyncio
import janus


##__________________________________________________________________||
class QueueDist:
    """Distribute data to subscribers

    Data can be sent from any thread. All subscribers need to be in the thread
    in which this class is instantiated.

    A new subscriber immediatly receives the most recent issue of the past data
    and then wait for future issues.

    The order of the data is preserved.
    """

    class End:
        pass

    class Start:
        pass

    def __init__(self):
        self._q_in = janus.Queue()

        self._qs_out = []  # list of janus.Queue()
        self._lock_out = threading.Condition()
        self._last_enumarated = (-1, self.Start)

        self._closed = False
        self._lock_close = asyncio.Condition()

        self._thread = threading.Thread(target=self._listen, daemon=True)
        self._thread.start()

    @property
    def nsubscriptions(self):
        """The number of the subscribers"""
        return len(self._qs_out)

    def put(self, item):
        """Send data to subscribers

        This method can be called in any thread.
        """
        self._q_in.sync_q.put(item)

    async def subscribe(self):
        """Asynchronous generator of data

        This method needs to be called in the thread in which this class
        is instantiated.
        """
        q = janus.Queue()

        with self._lock_out:
            self._qs_out.append(q)

        last_idx, last_item = self._last_enumarated

        if last_item is self.End:
            return

        try:
            if last_item is not self.Start:
                yield last_item

            while True:
                idx, item = await q.async_q.get()
                if item is self.End:
                    break
                if last_idx < idx:
                    yield item

        finally:
            with self._lock_out:
                self._qs_out.remove(q)

            q.close()
            await q.wait_closed()

    async def close(self):
        """End gracefully

        This method needs to be called in the thread in which this class
        is instantiated.
        """
        async with self._lock_close:
            if self._closed:
                return
            await self._close()
            self._closed = True

    async def _close(self):
        """Actual implementation of close()"""
        self._q_in.sync_q.put(self.End)

        try:
            await asyncio.to_thread(self._thread.join)
        except AttributeError:
            # for Python 3.8
            # to_thread() is new in Python 3.9
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, self._thread.join)

        self._q_in.close()
        await self._q_in.wait_closed()

    def _listen(self):
        """Distribution of data to subscribers

        This method runs in a thread.
        """
        idx, item = self._last_enumarated
        while item is not self.End:
            idx += 1
            item = self._q_in.sync_q.get()
            enumarated = (idx, item)
            with self._lock_out:
                for q in self._qs_out:
                    q.sync_q.put(enumarated)
            self._last_enumarated = enumarated


##__________________________________________________________________||



import queue
import asyncio
import warnings
import fnmatch

from .ci import PdbCommandInterface
from .custom import CustomizedPdb
from .stream import StreamIn, StreamOut

##__________________________________________________________________||
MODULES_TO_SKIP = [
    "threading",
    "queue",
    "importlib",
    "asyncio.*",
    "janus",
    "codec",
    "concurrent.futures.*",
    "selectors",
    "weakref",
    "_weakrefset",
    "socket",
    "logging",
    "os",
    "collections.*",
    "importlib.*",
    "pathlib",
    "typing",
    "posixpath",
    "fnmatch",
    "_pytest.*",
    "pluggy.*",
    "nextline.pdb.*",
    "nextline.queuedist",
    "nextlinegraphql.schema.bindables",
]


##__________________________________________________________________||
class PdbProxy:
    """A proxy of Pdb

    An instance of this class is created for each thread or async task.

    Parameters
    ----------
    thread_asynctask_id : object
        A thread and async tack ID
    trace : object
        A in stance of Trace
    modules_to_trace: set
        The set of modules to trace. This object is shared by multiple
        instances of this class. Modules in which Pdb commands are
        prompted will be added.
    registry: object
    ci_registry: object
    prompting_counter : callable
    """

    def __init__(
        self,
        thread_asynctask_id,
        trace,
        modules_to_trace,
        registry,
        ci_registry,
        prompting_counter,
    ):
        self.thread_asynctask_id = thread_asynctask_id
        self.trace = trace
        self.modules_to_trace = modules_to_trace
        self.registry = registry
        self.ci_registry = ci_registry
        self.skip = MODULES_TO_SKIP

        self.q_stdin = queue.Queue()
        self.q_stdout = queue.Queue()

        self.pdb = CustomizedPdb(
            proxy=self,
            prompting_counter=prompting_counter,
            stdin=StreamIn(self.q_stdin),
            stdout=StreamOut(self.q_stdout),
            skip=self.skip,
            readrc=False,
        )

        self._trace_func_all = self.trace_func_all
        self._traces = []

        self._first = True
        self._future = False

    def trace_func(self, frame, event, arg):
        """The main trace function

        This method will be called by the instance of Trace.
        The event should be always "call."
        """

        module_name = frame.f_globals.get("__name__")
        if self.pdb.is_skipped_module(module_name):
            return
        # print(module_name)

        if not event == "call":
            warnings.warn(
                f'The event is not "call": ({frame!r}, {event!r}, {arg!r})'
            )
        if self._first:
            return self.trace_func_register_new_thread_task(frame, event, arg)
        if self._future:
            return self.trace_func_reentry_after_future(frame, event, arg)
        return self.trace_func_all(frame, event, arg)

    def trace_func_register_new_thread_task(self, frame, event, arg):
        """The trace function for a new thread or async task

        The trace function of the first "call" event of the outermost
        scope of the thread or async task.

        """
        module_name = frame.f_globals.get("__name__")
        if not is_matched_to_any(module_name, self.modules_to_trace):
            return
        self._first = False
        self.registry.open_register(self.thread_asynctask_id)
        self.registry.register_list_item(
            "thread_task_ids", self.thread_asynctask_id
        )
        if self._trace_func_all:
            self._trace_func_all = self._trace_func_all(frame, event, arg)
        return self.trace_func_exit_thread_task

    def trace_func_reentry_after_future(self, frame, event, arg):
        """The trace function of reentry after "future" is returned"""
        module_name = frame.f_globals.get("__name__")
        if not is_matched_to_any(module_name, self.modules_to_trace):
            return
        self._future = False
        if self._trace_func_all:
            self._trace_func_all = self._trace_func_all(frame, event, arg)
        return self.trace_func_exit_thread_task

    def trace_func_exit_thread_task(self, frame, event, arg):
        """The trace function of the outermost scope in the thread or async task

        The trace function to detect the end of the thread or async
        task.

        """
        if self._trace_func_all:
            self._trace_func_all = self._trace_func_all(frame, event, arg)

        if event != "return":
            return self.trace_func_exit_thread_task

        if asyncio.isfuture(arg):
            # awaiting. will be called again
            self._future = True
            return

        self.trace.returning(self.thread_asynctask_id)
        self.registry.close_register(self.thread_asynctask_id)
        self.registry.deregister_list_item(
            "thread_task_ids", self.thread_asynctask_id
        )
        return

    def trace_func_all(self, frame, event, arg):
        """The trace function that calls the trace function of pdb"""

        module_name = frame.f_globals.get("__name__")
        # e.g., 'threading', '__main__', 'concurrent.futures.thread', 'asyncio.events'

        if self.pdb.is_skipped_module(module_name):
            # print(module_name)
            return

        func_name = frame.f_code.co_name
        # a function name
        # Note: '<module>' for the code produced by compile()
        if func_name == "<lambda>":
            return

        # print('{}.{}()'.format(module_name, func_name))
        # self.pdb.set_next(frame)

        trace = TraceBlock(
            thread_asynctask_id=self.thread_asynctask_id,
            pdb=self.pdb,
            registry=self.registry,
        )
        self._traces.append(trace)
        return trace(frame, event, arg)

    def entering_cmdloop(self, frame, state):
        """called by the customized pdb before it is entering the command loop"""
        module_name = frame.f_globals.get("__name__")
        self.modules_to_trace.add(module_name)

        self.pdb_ci = PdbCommandInterface(
            self.pdb, self.q_stdin, self.q_stdout
        )
        self.pdb_ci.start()
        self.ci_registry.add(self.thread_asynctask_id, self.pdb_ci)
        self.registry.register(self.thread_asynctask_id, state.copy())

    def exited_cmdloop(self, state):
        """called by the customized pdb after it has exited from the command loop"""
        self.ci_registry.remove(self.thread_asynctask_id)
        self.registry.register(self.thread_asynctask_id, state.copy())
        self.pdb_ci.end()


##__________________________________________________________________||
class TraceBlock:
    def __init__(self, thread_asynctask_id, pdb, registry):
        self.pdb = pdb
        self.trace_func = pdb.trace_dispatch
        self.registry = registry
        self.thread_asynctask_id = thread_asynctask_id

    def __call__(self, frame, event, arg):

        # if not frame.f_code.co_name == '<lambda>':
        #     file_name = self.pdb.canonic(frame.f_code.co_filename)
        #     line_no = frame.f_lineno
        #     self.registry.register_thread_task_state(self.thread_asynctask_id, file_name, line_no, event)

        if self.trace_func:
            self.trace_func = self.trace_func(frame, event, arg)
        return self


##__________________________________________________________________||
def is_matched_to_any(word, patterns):
    """
    based on Bdb.is_skipped_module()
    https://github.com/python/cpython/blob/v3.9.5/Lib/bdb.py#L191
    """
    if word is None:
        return False
    for pattern in patterns:
        if fnmatch.fnmatch(word, pattern):
            return True
    return False


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import GenericRepr, Snapshot


snapshots = Snapshot()

snapshots['test_args 1'] = [
    (
        'trace_func',
        (
            GenericRepr('sentinel.frame'),
            'call',
            None
        ),
        {
        }
    ),
    (
        'trace_func',
        (
            GenericRepr('sentinel.frame'),
            'line',
            None
        ),
        {
        }
    )
]



import sys
import threading


##__________________________________________________________________||
def exec_with_trace(code, trace, done=None):
    """execute code with trace

    Parameters
    ----------
    code : object
        A code to be executed. This must be an object that can be
        executed by the Python built-in Function exec().
    trace: callable
        A trace function.
    done: callable, optional
        A callable with two arguments. It will to be called after the
        code exits. The first argument is the return value, which is
        always None in the current implementation, because the
        built-in function exec() returns None. The second argument is
        the exception if an exception occurs or otherwise None.
    """

    ret = None
    exc = None

    globals_ = {"__name__": __name__}
    # To be given to exec() in order to address the issue
    # https://github.com/simonsobs/nextline/issues/7
    # __name__ is used in modules_to_trace in Trace.

    trace_org = sys.gettrace()
    threading.settrace(trace)
    sys.settrace(trace)
    try:
        ret = exec(code, globals_)
        # ret is always None
    except BaseException as e:
        exc = e
    finally:
        sys.settrace(trace_org)
        threading.settrace(trace_org)
        if done:
            done(ret, exc)


##__________________________________________________________________||



import asyncio


##__________________________________________________________________||
class ThreadSafeAsyncioEvent(asyncio.Event):
    """A thread-safe asyncio event

    The methods set() and clear() can be called from any threads. The method
    wait() needs to be called in the thread in which this class is instantiated.

    Code originally copied from https://stackoverflow.com/a/33006667/7309855

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._loop is None:  # Python 3.10
            self._loop = asyncio.get_event_loop()

    def set(self):
        return self._exec_threadsafe(super().set)

    def clear(self):
        return self._exec_threadsafe(super().clear)

    def _exec_threadsafe(self, func):
        if self._in_the_same_loop():
            return func()

        async def afunc(func):
            return func()

        fut = asyncio.run_coroutine_threadsafe(afunc(func), self._loop)
        return fut.result()

    def _in_the_same_loop(self):
        try:
            return self._loop is asyncio.get_running_loop()
        except RuntimeError:
            return False


##__________________________________________________________________||



import asyncio
from pathlib import Path

import pytest

from nextline import Nextline

##__________________________________________________________________||
statement = """
import script
script.run()

""".strip()


##__________________________________________________________________||
@pytest.fixture(autouse=True)
def monkey_patch_syspath(monkeypatch):
    this_dir = Path(__file__).resolve().parent
    monkeypatch.syspath_prepend(str(this_dir))
    yield


##__________________________________________________________________||
async def monitor_global_state(nextline):
    async for s in nextline.subscribe_global_state():
        print(s)


async def control_execution(nextline):
    controllers = {}
    async for ids in nextline.subscribe_thread_asynctask_ids():
        prev_ids = list(controllers.keys())
        new_ids = [id_ for id_ in ids if id_ not in prev_ids]
        ended_ids = [id_ for id_ in prev_ids if id_ not in ids]
        for id_ in new_ids:
            task = asyncio.create_task(control_thread_task(nextline, id_))
            controllers[id_] = task
        for id_ in ended_ids:
            del controllers[id_]


async def control_thread_task(nextline, thread_task_id):
    print(thread_task_id)
    async for s in nextline.subscribe_thread_asynctask_state(thread_task_id):
        print(s)
        if s["prompting"]:
            nextline.send_pdb_command(thread_task_id, "next")


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_run():

    nextline = Nextline(statement)

    assert nextline.global_state == "initialized"

    task_monitor_global_state = asyncio.create_task(
        monitor_global_state(nextline)
    )
    # await asyncio.sleep(0)

    task_control_execution = asyncio.create_task(control_execution(nextline))

    nextline.run()

    await nextline.finish()
    assert nextline.global_state == "finished"
    await nextline.close()
    assert nextline.global_state == "closed"


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_proxy[asyncio] 1'] = [
    [
        (
            'run_a',
            'call',
            False
        ),
        (
            'a',
            'call',
            False
        ),
        (
            'a',
            'call',
            False
        )
    ],
    [
        (
            'run_a',
            'line',
            False
        ),
        (
            'a',
            'line',
            False
        ),
        (
            'a',
            'exception',
            False
        )
    ],
    [
        (
            'a',
            'return',
            True
        ),
        (
            'a',
            'line',
            False
        ),
        (
            'run_a',
            'return',
            False
        )
    ],
    [
        (
            'a',
            'return',
            False
        )
    ]
]

snapshots['test_proxy[lambda] 1'] = [
    [
        (
            'call_lambda',
            'call',
            False
        )
    ],
    [
        (
            'call_lambda',
            'line',
            False
        )
    ],
    [
        (
            'call_lambda',
            'line',
            False
        )
    ],
    [
        (
            'call_lambda',
            'return',
            False
        )
    ]
]

snapshots['test_proxy[nested-func] 1'] = [
    [
        (
            'subject',
            'call',
            False
        ),
        (
            'f',
            'call',
            False
        ),
        (
            'f',
            'call',
            False
        )
    ],
    [
        (
            'subject',
            'line',
            False
        ),
        (
            'f',
            'line',
            False
        ),
        (
            'f',
            'line',
            False
        )
    ],
    [
        (
            'f',
            'line',
            False
        ),
        (
            'subject',
            'line',
            False
        ),
        (
            'f',
            'line',
            False
        )
    ],
    [
        (
            'f',
            'return',
            False
        ),
        (
            'f',
            'return',
            False
        ),
        (
            'subject',
            'line',
            False
        )
    ],
    [
        (
            'subject',
            'return',
            False
        )
    ]
]

snapshots['test_proxy[simple] 1'] = [
    [
        (
            'f',
            'call',
            False
        )
    ],
    [
        (
            'f',
            'line',
            False
        )
    ],
    [
        (
            'f',
            'line',
            False
        )
    ],
    [
        (
            'f',
            'return',
            False
        )
    ]
]

snapshots['test_proxy[yield] 1'] = [
    [
        (
            'call_gen',
            'call',
            False
        ),
        (
            'gen',
            'call',
            False
        ),
        (
            'gen',
            'call',
            False
        ),
        (
            'gen',
            'call',
            False
        )
    ],
    [
        (
            'call_gen',
            'line',
            False
        ),
        (
            'gen',
            'line',
            False
        ),
        (
            'gen',
            'line',
            False
        ),
        (
            'gen',
            'return',
            False
        )
    ],
    [
        (
            'gen',
            'return',
            False
        ),
        (
            'call_gen',
            'line',
            False
        ),
        (
            'gen',
            'return',
            False
        )
    ],
    [
        (
            'call_gen',
            'line',
            False
        )
    ],
    [
        (
            'call_gen',
            'line',
            False
        )
    ],
    [
        (
            'call_gen',
            'line',
            False
        )
    ],
    [
        (
            'call_gen',
            'exception',
            False
        )
    ],
    [
        (
            'call_gen',
            'return',
            False
        )
    ]
]



import asyncio
from typing import Coroutine, Awaitable, Union


##__________________________________________________________________||
class CoroutineRunner:
    """Execute the coroutine in the asyncio event loop at the instantiation

    A running asyncio event loop needs to exist in the thread in which this
    class is instantiated. The method run() can be called in any thread,
    regardless of whether the same thread in which this class is instantiated or
    any other thread. In either case, the coroutine given to run() will be
    executed in the event loop in the thread in which this class is
    instantiated.

    This class can be useful when you need to run coroutines in the asyncio
    event loop in a particular thread but events that require the coroutines to
    run can occur any threads.
    """

    def __init__(self):
        self.loop = asyncio.get_running_loop()

    def __repr__(self):
        # e.g., "<CoroutineRunner loop=<_UnixSelectorEventLoop running=True closed=False debug=False>>"
        return f"<{self.__class__.__name__} loop={self.loop!r}>"

    def run(self, coro: Coroutine) -> Union[Awaitable, None]:
        """Schedule or execute the coroutine

        If this method is called in the same thread as the one in which this
        class was instantizted, return the asyncio task for the coroutine. If in
        a different thread, schedule to run in the thread at the instantiation,
        wait until it exits, and return None.

        Parameters
        ----------
        coro : Coroutine
            The coroutine to be executed

        Returns
        -------
        Task or None
            The task for the coroutine if in the same thread as of the
            instantiation of this class. None, otherwise.
        """

        if self._is_the_same_running_loop():
            return asyncio.create_task(coro)

        if self.loop.is_closed():
            # The loop that was running when this class
            # was inntantizted is closed.
            raise RuntimeError(f"The loop is closed: {self.loop}")

        # In another thread

        fut = asyncio.run_coroutine_threadsafe(coro, self.loop)
        fut.result()

    def _is_the_same_running_loop(self):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            return False
        return self.loop is loop


##__________________________________________________________________||



import asyncio
import threading
from typing import Optional, Union

from .trace import Trace
from .utils import Registry, ThreadSafeAsyncioEvent
from .exec_ import exec_with_trace

SCRIPT_FILE_NAME = "<string>"


# __________________________________________________________________||
class Machine:
    """State machine

                 .-------------.
            .--->| Initialized |---.
    reset() |    '-------------'   |
            |      |   | run()     |
            |------'   |           |
            |          v           |
            |    .-------------.   |
            |    |   Running   |   |
            |    '-------------'   |
            |          | _exited() |
            |          |           |
            |          v           |
            |    .-------------.   |
            |    |   Exited    |   |
            |    '-------------'   |
            |          | finish()  |
            |          |           |
            |          V           |
            |    .-------------.   |
            |----|   Finished  |   |
            |    '-------------'   |
            |          |-----------'
            |          | close()
            |          V
            |    .-------------.
            '----|   Closed    |
                 '-------------'

    """

    def __init__(self, statement):
        self.condition_finish = asyncio.Condition()
        self.condition_close = asyncio.Condition()

        self.state = Initialized(statement)
        self.registry = self.state.registry

    def __repr__(self):
        # e.g., "<Machine 'running'>"
        return f"<{self.__class__.__name__} {self.state_name!r}>"

    @property
    def state_name(self) -> str:
        """e.g., "initialized", "running","""
        return self.state.name

    def run(self):
        """Enter the running state"""
        self.state = self.state.run()
        self._task_exited = asyncio.create_task(self._exited())

    async def _exited(self):
        """Enter the exited state

        This method is scheduled as a task in run().

        Bofore changing the state to "exited", this method waits for the script
        execution (in another thread) to exit.
        """

        self.state = await self.state.exited()

    def send_pdb_command(self, thread_asynctask_id, command):
        self.state.send_pdb_command(thread_asynctask_id, command)

    async def finish(self):
        """Enter the finished state"""
        await self._task_exited
        async with self.condition_finish:
            self.state = await self.state.finish()

    def exception(self):
        return self.state.exception()

    def result(self):
        self.state.result()

    def reset(self, statement=None):
        """Enter the initialized state"""
        self.state = self.state.reset(statement=statement)

    async def close(self):
        """Enter the closed state"""
        async with self.condition_close:
            self.state = await self.state.close()


# __________________________________________________________________||
class StateObsoleteError(Exception):
    """Operation on an obsolete state object."""

    pass


class StateMethodError(Exception):
    """Irrelevant operation on the state."""

    pass


# __________________________________________________________________||
class ObsoleteMixin:
    def assert_not_obsolete(self):
        if self.is_obsolete():
            raise StateObsoleteError(f"The state object is obsolete: {self!r}")

    def is_obsolete(self):
        return getattr(self, "_obsolete", False)

    def obsolete(self):
        self._obsolete = True


class State(ObsoleteMixin):
    """The base state class in the Nextline state machine"""

    def __repr__(self):
        # e.g., "<Initialized 'initialized'>"
        items = [self.__class__.__name__, repr(self.name)]
        if self.is_obsolete():
            items.append("obsolete")
        return f'<{" ".join(items)}>'

    def run(self):
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def exited(self):
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def finish(self):
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def reset(self, *_, **__):
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def close(self):
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def send_pdb_command(self, *_, **__):
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def exception(self):
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def result(self):
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")


class Initialized(State):
    """The state "initialized", ready to run

    Parameters
    ----------
    statement : str or None
        A Python code as a string. None can be given by a state object
        when the state is reset. If None, the statement will be
        obtained from the registry.
    registry : object, optional
        An instance of Registry. This parameter will be given by a state
        object when the state is reset.

    """

    name = "initialized"

    def __init__(
        self, statement: Union[str, None], registry: Optional[Registry] = None
    ):

        if registry:
            self.registry = registry
        else:
            self.registry = Registry()
            self.registry.open_register("statement")
            self.registry.open_register("state_name")
            self.registry.open_register("script_file_name")
            self.registry.open_register_list("thread_task_ids")

        if statement:
            self.registry.register("statement", statement)
            self.registry.register("script_file_name", SCRIPT_FILE_NAME)
        else:
            statement = self.registry.get("statement")

        if isinstance(statement, str):
            self._code = compile(statement, SCRIPT_FILE_NAME, "exec")
        else:
            self._code = statement

        self.registry.register("state_name", self.name)

    def run(self):
        self.assert_not_obsolete()
        running = Running(self.registry, self._code)
        self.obsolete()
        return running

    def reset(self, statement=None):
        self.assert_not_obsolete()
        initialized = Initialized(statement=statement, registry=self.registry)
        self.obsolete()
        return initialized

    async def close(self):
        self.assert_not_obsolete()
        closed = Closed(self.registry)
        await closed._ainit()
        self.obsolete()
        return closed


class Running(State):
    """The state "running", the script is being executed.

    Parameters
    ----------
    registry : object
        An instance of Registry
    code : object
        The code object to be executed.
    """

    name = "running"

    def __init__(self, registry, code):
        self.registry = registry
        self._event = ThreadSafeAsyncioEvent()

        trace = Trace(
            registry=self.registry,
            modules_to_trace={exec_with_trace.__module__},
        )
        self.pdb_ci_registry = trace.pdb_ci_registry

        self.registry.register("state_name", self.name)

        self.loop = asyncio.get_running_loop()

        self._thread = threading.Thread(
            target=exec_with_trace, args=(code, trace, self._done), daemon=True
        )
        self._thread.start()

    def _done(self, result=None, exception=None):
        # callback function, to be called from another thread at the
        # end of exec_with_trace()

        if self.loop.is_closed():
            # The exit is not being waited in the main thread, for
            # example, neither exited() of finish() is called.
            return

        self._exited = Exited(
            self.registry,
            thread=self._thread,
            result=result,
            exception=exception,
        )

        self.obsolete()

        try:
            self._event.set()
        except RuntimeError:
            # The event loop is closed.
            pass

    async def exited(self):
        """return the exited state after the script exits."""
        await self._event.wait()
        return self._exited

    def send_pdb_command(self, thread_asynctask_id, command):
        pdb_ci = self.pdb_ci_registry.get_ci(thread_asynctask_id)
        pdb_ci.send_pdb_command(command)


class Exited(State):
    """The state "exited", the script execution has exited

    Parameters
    ----------
    registry : object
        An instance of Registry
    thread : object
        The object of the thread in which the script was executed.
        This thread is to be joined.
    result : any
        The result of the script execution, always None
    exception : exception or None
        The execution raised in the script execution if any. Otherwise
        None
    """

    name = "exited"

    def __init__(self, registry, thread, result, exception):
        self.registry = registry
        self._thread = thread
        self._result = result
        self._exception = exception

        self.registry.register("state_name", self.name)

    async def finish(self):
        self.assert_not_obsolete()
        await self._join(self._thread)
        finished = Finished(
            self.registry, result=self._result, exception=self._exception
        )
        self.obsolete()
        return finished

    async def _join(self, thread):
        try:
            await asyncio.to_thread(thread.join)
        except AttributeError:
            # for Python 3.8
            # to_thread() is new in Python 3.9
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, thread.join)


class Finished(State):
    """The state "finished", the script execution has finished

    The thread which executed the script has been joined.

    Parameters
    ----------
    registry : object
        An instance of Registry
    result : any
        The result of the script execution, always None
    exception : exception or None
        The exception of the script execution if any. Otherwise None

    """

    name = "finished"

    def __init__(self, registry, result, exception):
        self._result = result
        self._exception = exception

        self.registry = registry
        self.registry.register("state_name", self.name)

    def exception(self):
        """Return the exception of the script execution

        Return None if no execution has been raised.

        """
        return self._exception

    def result(self):
        """Return the result of the script execution

        None in the current implementation as the build-in function
        exec() returns None.

        Re-raise the exception if an exception has been raised in the
        script execution.

        """

        if self._exception:
            raise self._exception

        return self._result

    async def finish(self):
        # This method can be called when Nextline.finish() is
        # asynchronously called multiple times.
        self.assert_not_obsolete()
        return self

    def reset(self, statement=None):
        self.assert_not_obsolete()
        initialized = Initialized(statement=statement, registry=self.registry)
        self.obsolete()
        return initialized

    async def close(self):
        self.assert_not_obsolete()
        closed = Closed(self.registry)
        await closed._ainit()
        self.obsolete()
        return closed


class Closed(State):
    """The state "closed"

    Parameters
    ----------
    registry : object
        An instance of Registry
    """

    name = "closed"

    def __init__(self, registry):
        self.registry = registry
        self.statement = self.registry.get("statement")
        self.registry.register("state_name", self.name)

    async def _ainit(self):
        await self.registry.close()  # close here because "await" is
        # not allowed in __init__()

    def reset(self, statement=None):
        self.assert_not_obsolete()
        if statement is None:
            statement = self.statement
        initialized = Initialized(statement=statement)
        self.obsolete()
        return initialized

    async def close(self):
        # This method can be called when Nextline.close() is
        # asynchronously called multiple times.
        self.assert_not_obsolete()
        return self


# __________________________________________________________________||



import sys
import asyncio
import threading
from functools import partial

import pytest

from nextline.utils import UniqThreadTaskIdComposer


##__________________________________________________________________||
@pytest.fixture(autouse=True)
def wrap_thread(monkeypatch):
    """Wrap threading.Thread so to raise exception in the caller

    The implementation based on
    https://www.geeksforgeeks.org/handling-a-threads-exception-in-the-caller-thread-in-python/

    """

    import threading

    class Wrapped(threading.Thread):
        def run(self):
            self.exc = None
            try:
                return super().run()
            except BaseException as e:
                self.exc = e

        def join(self):
            ret = super().join()
            if self.exc:
                raise self.exc
            return ret

    monkeypatch.setattr(threading, "Thread", Wrapped)
    yield


##__________________________________________________________________||
def assert_func(obj, expected):
    assert expected == obj.compose()
    assert expected == obj.compose()


async def async_assert_func(obj, expected):
    return assert_func(obj, expected)


##__________________________________________________________________||
@pytest.fixture()
def obj():
    y = UniqThreadTaskIdComposer()
    yield y


##__________________________________________________________________||
def test_compose(obj):
    expected = (1, None)
    assert_func(obj, expected)


def test_threads(obj):
    expected = (1, None)
    assert_func(obj, expected)

    expected = (2, None)
    t = threading.Thread(target=assert_func, args=(obj, expected))
    t.start()
    t.join()
    obj.exited(expected)

    expected = (3, None)
    t = threading.Thread(target=assert_func, args=(obj, expected))
    t.start()
    t.join()
    obj.exited(expected)


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_async_coroutine(obj):
    expected = (1, 1)
    assert_func(obj, expected)

    # run in the same task
    await async_assert_func(obj, expected)
    await async_assert_func(obj, expected)


@pytest.mark.asyncio
async def test_async_tasks(obj):
    expected = (1, 1)
    assert_func(obj, expected)

    expected = (1, 2)
    t = asyncio.create_task(async_assert_func(obj, expected))
    await t
    obj.exited(expected)

    expected = (1, 3)
    t = asyncio.create_task(async_assert_func(obj, expected))
    await t
    obj.exited(expected)


@pytest.mark.asyncio
async def test_async_tasks_gather(obj):
    expected = (1, 1)
    assert_func(obj, expected)

    expected = (1, 2)
    t1 = asyncio.create_task(async_assert_func(obj, expected))
    expected = (1, 3)
    t2 = asyncio.create_task(async_assert_func(obj, expected))
    aws = {t1, t2}
    await asyncio.gather(*aws)


def test_async_asyncio_run(obj):
    expected = (1, None)
    assert_func(obj, expected)

    expected = (1, 1)
    asyncio.run(async_assert_func(obj, expected))


@pytest.mark.skipif(sys.version_info < (3, 9), reason="asyncio.to_thread()")
@pytest.mark.asyncio
async def test_async_asyncio_to_thread(obj):
    expected = (1, 1)
    assert_func(obj, expected)

    expected = (2, None)
    await asyncio.to_thread(partial(assert_func, obj, expected))
    obj.exited(expected)

    expected = (3, None)
    await asyncio.to_thread(partial(assert_func, obj, expected))
    obj.exited(expected)


##__________________________________________________________________||
async def async_nested(obj, expected_thread_id):
    expected = (expected_thread_id, 1)
    assert_func(obj, expected)
    await async_assert_func(obj, expected)

    expected = (expected_thread_id, 2)
    t1 = asyncio.create_task(async_assert_func(obj, expected))
    expected = (expected_thread_id, 3)
    t2 = asyncio.create_task(async_assert_func(obj, expected))
    aws = {t1, t2}
    await asyncio.gather(*aws)
    obj.exited((expected_thread_id, 2))
    obj.exited((expected_thread_id, 3))


def nested(obj, expected_thread_id):
    expected = (expected_thread_id, None)
    assert_func(obj, expected)

    asyncio.run(async_nested(obj, expected_thread_id))
    obj.exited((expected_thread_id, 1))


def test_nested(obj):
    expected = (1, None)
    assert_func(obj, expected)

    expected_thread_id = 2
    t = threading.Thread(target=nested, args=(obj, expected_thread_id))
    t.start()
    t.join()
    obj.exited((expected_thread_id, None))

    expected_thread_id = 3
    t = threading.Thread(target=nested, args=(obj, expected_thread_id))
    t.start()
    t.join()
    obj.exited((expected_thread_id, None))


##__________________________________________________________________||



import threading
import asyncio
import warnings
from typing import Hashable, Coroutine

from .coro_runner import CoroutineRunner
from .queuedist import QueueDist


##__________________________________________________________________||
class Registry:
    """Subscribable asynchronous thread-safe registers"""

    def __init__(self):
        self._runner = CoroutineRunner()
        self._lock = threading.Condition()
        self._data = {}
        self._queue = {}
        self._aws = []

    async def close(self):
        """End gracefully"""
        if self._aws:
            await asyncio.gather(*self._aws)
        while self._queue:
            _, q = self._queue.popitem()
            await q.close()
        self._data.clear()

    def open_register(self, key: Hashable):
        """Create a register for an item"""
        self._open_register(key, None)

    def open_register_list(self, key: Hashable):
        """Create a register for an item list"""
        self._open_register(key, [])

    def _open_register(self, key, init_data):
        if key in self._data:
            raise Exception(f"register key already exists {key!r}")
        self._data[key] = init_data
        self._create_queue_from_another_thread(key)

    def close_register(self, key: Hashable):
        try:
            del self._data[key]
        except KeyError:
            warnings.warn(f"key not found: {key}")
        self._close_queue_from_another_thread(key)

    def register(self, key, item):
        """Replace the item in the register"""
        if key not in self._data:
            raise Exception(f"register key does not exist {key!r}")

        self._data[key] = item

        self._distribute(key, item)

    def register_list_item(self, key, item):
        """Add an item to the register"""
        if key not in self._data:
            raise Exception(f"register key does not exist {key!r}")

        with self._lock:
            self._data[key].append(item)
            copy = self._data[key].copy()

        self._distribute(key, copy)

    def deregister_list_item(self, key, item):
        """Remove the item from the register"""
        if key not in self._data:
            raise Exception(f"register key does not exist {key!r}")

        with self._lock:
            try:
                self._data[key].remove(item)
            except ValueError:
                warnings.warn(f"item not found: {item}")
            copy = self._data[key].copy()

        self._distribute(key, copy)

    def get(self, key):
        """The item in the register"""
        return self._data[key]

    async def subscribe(self, key):
        """Asynchronous generator of items in the register

        This method needs to be called in the initial thread.
        """
        queue = self._queue.get(key)
        if not queue:
            self._create_queue(key)
            queue = self._queue.get(key)
        async for y in queue.subscribe():
            yield y

    def _create_queue_from_another_thread(self, key):
        """Open a queue

        Can be called from any threads
        """

        async def create():
            self._create_queue(key)

        self._run_in_the_initial_thread(create())

    def _create_queue(self, key):
        """Open a queue

        To be run in the initial thread
        """
        if key in self._queue:
            return
        self._queue[key] = QueueDist()

    def _close_queue_from_another_thread(self, key):
        """Remove the queue

        Can be called from any threads
        """
        coro = self._close_queue(key)
        self._run_in_the_initial_thread(coro)

    async def _close_queue(self, key):
        """Remove the queue

        To be run in the initial thread
        """
        queue = self._queue.pop(key, None)
        if queue:
            await queue.close()

    def _distribute(self, key, item):
        """Send item to subscribers"""

        async def put():
            self._queue[key].put(item)

        self._run_in_the_initial_thread(put())

    def _run_in_the_initial_thread(self, coro: Coroutine):
        """Schedule the coroutine as an asyncio task

        The task will be scheduled to the asyncio event loop in the thread in
        which this class is instantizted.
        """
        task = self._runner.run(coro)
        if task:
            self._aws.append(task)


##__________________________________________________________________||



import threading
import asyncio
from collections import namedtuple
from itertools import count


##__________________________________________________________________||
class UniqThreadTaskIdComposer:
    """Compose paris of unique thread Id and async task Id"""

    ThreadSpecifics = namedtuple(
        "ThreadSpecifics", ["thread_ident", "task_id_counter", "task_ids"]
    )
    # thread_ident: threading.get_ident()
    # task_id_counter: count().__next__
    # task_ids: set()

    def __init__(self):

        self.non_uniq_id_dict = {}  # non_uniq_id -> uniq_id
        self.uniq_id_dict = {}  # uniq_id -> non_uniq_id
        # uniq_id = (thread_id, task_id)
        # non_uniq_id = (threading.get_ident(), id(asyncio.current_task()))

        self.non_uniq_thread_id_dict = {}  # threading.get_ident() -> thread_id
        self.thread_id_dict = {}  # thread_id -> ThreadSpecifics

        self.thread_id_counter = count().__next__
        self.thread_id_counter()  # consume 0

        self.condition = threading.Condition()

    def compose(self):
        """Return the pair of the current thread ID and async task ID

        Returns
        -------
        tuple
            The pair of the current thread ID and async task ID. If
            not in an async task, the async task ID will be None.
        """

        non_uniq_id = self._compose_non_uniq_id()

        uniq_id = self._find_previsouly_composed_uniq_id(non_uniq_id)
        if uniq_id:
            return uniq_id

        return self._compose_uniq_id(non_uniq_id)

    def exited(self, thread_task_id):
        thread_id, task_id = thread_task_id
        with self.condition:
            non_uniq_thread_id, non_uniq_task_id = self.uniq_id_dict.pop(
                thread_task_id
            )
            self.non_uniq_id_dict.pop((non_uniq_thread_id, non_uniq_task_id))
            self.thread_id_dict[thread_id].task_ids.remove(task_id)
            if not self.thread_id_dict[thread_id].task_ids:
                self.non_uniq_thread_id_dict.pop(non_uniq_thread_id)

    def _compose_non_uniq_id(self):

        non_uniq_thread_id = threading.get_ident()
        # the "thread identifier", which can be recycled after a thread exits
        # https://docs.python.org/3/library/threading.html#threading.get_ident

        non_uniq_task_id = None
        try:
            non_uniq_task_id = id(asyncio.current_task())
            # the id of the task object, which can be also recycled
            # https://docs.python.org/3/library/functions.html#id
        except RuntimeError:
            # no running event loop
            pass

        return non_uniq_thread_id, non_uniq_task_id

    def _find_previsouly_composed_uniq_id(self, non_uniq_id):
        try:
            return self.non_uniq_id_dict[non_uniq_id]
        except KeyError:
            return None

    def _compose_uniq_id(self, non_uniq_id):
        thread_id = self._find_previsouly_composed_thread_id(non_uniq_id)
        if not thread_id:
            thread_id = self._compose_thread_id(non_uniq_id)
        task_id = self._compose_task_id(non_uniq_id)
        uniq_id = (thread_id, task_id)
        with self.condition:
            self.non_uniq_id_dict[non_uniq_id] = uniq_id
            self.uniq_id_dict[uniq_id] = non_uniq_id
        return uniq_id

    def _find_previsouly_composed_thread_id(self, non_uniq_id):
        non_uniq_thread_id = non_uniq_id[0]
        try:
            return self.non_uniq_thread_id_dict[non_uniq_thread_id]
        except KeyError:
            return None

    def _compose_thread_id(self, non_uniq_id):

        non_uniq_thread_id = non_uniq_id[0]
        thread_id = self.thread_id_counter()

        task_id_counter = count().__next__
        task_id_counter()  # consume 0

        thread_specifics = self.ThreadSpecifics(
            thread_ident=non_uniq_thread_id,
            task_id_counter=task_id_counter,
            task_ids=set(),
        )

        with self.condition:
            self.non_uniq_thread_id_dict[non_uniq_thread_id] = thread_id
            self.thread_id_dict[thread_id] = thread_specifics
        return thread_id

    def _compose_task_id(self, non_uniq_id):

        non_uniq_thread_id, non_uniq_task_id = non_uniq_id

        thread_id = self.non_uniq_thread_id_dict[non_uniq_thread_id]

        thread_specifics = self.thread_id_dict[thread_id]

        task_id = None
        if non_uniq_task_id:
            task_id = thread_specifics.task_id_counter()
        thread_specifics.task_ids.add(task_id)

        return task_id


##__________________________________________________________________||
2



import sys
import asyncio
from itertools import count

import pytest
from unittest.mock import Mock

from nextline.trace import Trace
from nextline.utils import Registry
from nextline.pdb.proxy import PdbProxy
from nextline.trace import UniqThreadTaskIdComposer

from . import subject


##__________________________________________________________________||
@pytest.fixture()
def mock_trace():
    y = Mock(spec=Trace)
    yield y


@pytest.fixture()
def mock_registry():
    y = Mock(spec=Registry)
    yield y


@pytest.fixture()
def proxy(mock_trace, mock_registry):
    thread_asynctask_id = UniqThreadTaskIdComposer().compose()

    modules_to_trace = {"tests.pdb.subject"}

    prompting_counter = count().__next__
    prompting_counter()  # consume 0

    y = PdbProxy(
        thread_asynctask_id=thread_asynctask_id,
        trace=mock_trace,
        modules_to_trace=modules_to_trace,
        registry=mock_registry,
        ci_registry=Mock(),
        prompting_counter=prompting_counter,
    )

    y.pdb.trace_dispatch = Mock()

    yield y


##__________________________________________________________________||
def unpack_trace_dispatch_call(trace_dispatch):
    trace_results = []
    while trace_dispatch.call_count:
        trace_results.append(
            [
                (
                    c.args[0].f_code.co_name,
                    c.args[1],
                    asyncio.isfuture(c.args[2]),
                )
                for c in trace_dispatch.call_args_list
            ]
        )
        trace_dispatch = trace_dispatch.return_value

    # e.g.,
    # trace_results = [
    #     [('run_a', 'call', False), ('<lambda>', 'call', False), ('a', 'call', False), ('a', 'call', False)],
    #     [('run_a', 'line', False), ('<lambda>', 'line', False), ('a', 'line', False), ('a', 'exception', False)],
    #     [('<lambda>', 'return', False), ('a', 'return', True), ('a', 'line', False), ('run_a', 'return', False)],
    #     [('a', 'return', False)]
    # ]

    return trace_results


##__________________________________________________________________||
params = [
    pytest.param(subject.f, id="simple"),
    pytest.param(subject.subject, id="nested-func"),
    pytest.param(
        subject.call_gen,
        id="yield",
        marks=pytest.mark.skipif(
            sys.version_info >= (3, 10), reason="StopIteration won't be raised"
        ),
    ),
    pytest.param(subject.run_a, id="asyncio"),
    pytest.param(subject.call_lambda, id="lambda"),
]


@pytest.mark.skipif(
    sys.version_info < (3, 9), reason="co_name <lambda> is different "
)
@pytest.mark.parametrize("subject", params)
def test_proxy(proxy, mock_trace, mock_registry, snapshot, subject):
    """test PdbProxy"""
    # TODO: the test needs to be restructured so that, for example, a
    # coroutine or a generator can be the outermost scope.

    trace_org = sys.gettrace()
    sys.settrace(proxy.trace_func)
    subject()
    sys.settrace(trace_org)

    assert 1 == mock_registry.open_register.call_count
    assert 1 == mock_registry.register_list_item.call_count
    assert 1 == mock_registry.close_register.call_count

    assert 1 == mock_trace.returning.call_count

    trace_results = unpack_trace_dispatch_call(proxy.pdb.trace_dispatch)
    snapshot.assert_match(trace_results)


##__________________________________________________________________||



import pytest

from nextline import Nextline

##__________________________________________________________________||
statement = """
import time
time.sleep(0.01)
"""


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_run():

    nextline = Nextline(statement)
    assert nextline.global_state == "initialized"

    nextline.run()

    async for global_state in nextline.subscribe_global_state():
        if global_state == "running":
            break

    async for thread_asynctask_ids in nextline.subscribe_thread_asynctask_ids():
        if thread_asynctask_ids:
            thread_asynctask_id = thread_asynctask_ids[0]
            break

    async for state in nextline.subscribe_thread_asynctask_state(
        thread_asynctask_id
    ):
        if state["prompting"]:
            break

    nextline.send_pdb_command(thread_asynctask_id, "continue")
    await nextline.finish()
    assert nextline.global_state == "finished"
    await nextline.close()
    assert nextline.global_state == "closed"


##__________________________________________________________________||



import sys
import asyncio

import pytest

from nextline.utils import Registry

from .aiterable import aiterable


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_simple_usage():
    obj = Registry()

    key = "key_one"
    items = ["item_one", "item_two", "item_three"]

    async def subscribe():
        return [y async for y in obj.subscribe(key)]

    async def register():
        obj.open_register(key)
        async for i in aiterable(items):
            obj.register(key, i)
        obj.close_register(key)

    results = await asyncio.gather(subscribe(), register())
    assert items == results[0]

    await obj.close()


@pytest.mark.asyncio
async def test_simple_usage_list():
    obj = Registry()

    key = "key_one"
    items = ["item_one", "item_two", "item_three"]
    expected = [
        ["item_one"],
        ["item_one", "item_two"],
        ["item_one", "item_two", "item_three"],
        ["item_two", "item_three"],
        ["item_three"],
        [],
    ]

    async def subscribe():
        return [y async for y in obj.subscribe(key)]

    async def register():
        obj.open_register_list(key)
        async for i in aiterable(items):
            obj.register_list_item(key, i)
        async for i in aiterable(items):
            obj.deregister_list_item(key, i)
        obj.close_register(key)

    results = await asyncio.gather(subscribe(), register())
    assert expected == results[0]

    await obj.close()


##__________________________________________________________________||
nitems = [0, 1, 2, 50]
nsubscribers = [0, 1, 2, 70]


@pytest.mark.skipif(sys.version_info < (3, 9), reason="asyncio.to_thread() ")
@pytest.mark.parametrize("nitems", nitems)
@pytest.mark.parametrize("nsubscribers", nsubscribers)
@pytest.mark.parametrize("thread", [True, False])
@pytest.mark.parametrize("close_register", [True, False])
@pytest.mark.asyncio
async def test_one(close_register, thread, nsubscribers, nitems):
    async def subscribe(registry, register_key):
        ret = []
        async for y in registry.subscribe(register_key):
            ret.append(y)
        return ret

    def register(registry, register_key, items):
        registry.open_register(register_key)
        assert registry.get(register_key) is None
        for item in items:
            registry.register(register_key, item)
            assert registry.get(register_key) == item
        if close_register:
            registry.close_register(register_key)

    async def aregister(registry, register_key, items):
        return register(registry, register_key, items)

    registry = Registry()

    register_key = "item"
    items = [f"{register_key}-{i+1}" for i in range(nitems)]

    # subscribe
    tasks_subscribe = []
    for i in range(nsubscribers):
        task = asyncio.create_task(subscribe(registry, register_key))
        tasks_subscribe.append(task)

    # register
    if thread:
        coro = asyncio.to_thread(register, registry, register_key, items)
    else:
        coro = aregister(registry, register_key, items)
    task_register = asyncio.create_task(coro)

    await asyncio.gather(task_register)

    task_close = asyncio.create_task(registry.close())

    results = await asyncio.gather(*tasks_subscribe, task_close)
    expected = items
    actuals = results[:nsubscribers]
    for actual in actuals:
        assert actual == expected


##__________________________________________________________________||
nitems = [0, 1, 2, 10]
nsubscribers = [0, 1, 2, 70]


@pytest.mark.skipif(sys.version_info < (3, 9), reason="asyncio.to_thread() ")
@pytest.mark.parametrize("nitems", nitems)
@pytest.mark.parametrize("nsubscribers", nsubscribers)
@pytest.mark.parametrize("thread", [True, False])
@pytest.mark.parametrize("close_register", [True, False])
@pytest.mark.asyncio
async def test_list(close_register, thread, nsubscribers, nitems):
    async def subscribe(registry, register_key):
        ret = []
        async for y in registry.subscribe(register_key):
            ret.append(y)
        return ret

    def register(registry, register_key, items):
        registry.open_register_list(register_key)
        assert registry.get(register_key) == []
        for i, item in enumerate(items):
            registry.register_list_item(register_key, item)
            assert registry.get(register_key) == items[: i + 1]
        for i, item in enumerate(items):
            registry.deregister_list_item(register_key, item)
            assert registry.get(register_key) == items[i + 1 :]
        if close_register:
            registry.close_register(register_key)

    async def aregister(registry, register_key, items):
        return register(registry, register_key, items)

    registry = Registry()

    register_key = "item"
    items = [f"{register_key}-{i+1}" for i in range(nitems)]

    # subscribe
    tasks_subscribe = []
    for i in range(nsubscribers):
        task = asyncio.create_task(subscribe(registry, register_key))
        tasks_subscribe.append(task)

    # register
    if thread:
        coro = asyncio.to_thread(register, registry, register_key, items)
    else:
        coro = aregister(registry, register_key, items)
    task_register = asyncio.create_task(coro)

    await asyncio.gather(task_register)

    task_close = asyncio.create_task(registry.close())

    results = await asyncio.gather(*tasks_subscribe, task_close)
    expected = [items[: i + 1] for i in range(len(items))]
    expected.extend([items[i + 1 :] for i in range(len(items))])
    actuals = results[:nsubscribers]
    for actual in actuals:
        assert actual == expected


##__________________________________________________________________||



from itertools import count

from .pdb.proxy import PdbProxy
from .registry import PdbCIRegistry
from .utils import UniqThreadTaskIdComposer


##__________________________________________________________________||
class Trace:
    """The main trace function

    An instance of this class, which is callable, should be set as the
    trace function by sys.settrace() and threading.settrace().

    Parameters
    ----------
    registry : object
        An instance of Registry
    modules_to_trace : set
        The names of modules to trace
    """

    def __init__(self, registry, modules_to_trace):

        self.registry = registry
        self.pdb_ci_registry = PdbCIRegistry()

        self.prompting_counter = count().__next__
        self.prompting_counter()  # consume 0

        self.pdb_proxies = {}

        self.modules_to_trace = set(modules_to_trace)
        # Make a copy so that the original won't be modified.
        # self.modules_to_trace will be shared and modified by
        # multiple instances of PdbProxy.

        self.id_composer = UniqThreadTaskIdComposer()

    def __call__(self, frame, event, arg):
        """Called by the Python interpreter when a new local scope is entered.

        https://docs.python.org/3/library/sys.html#sys.settrace

        """

        thread_asynctask_id = self.id_composer.compose()
        # print(*thread_asynctask_id)

        pdb_proxy = self.pdb_proxies.get(thread_asynctask_id)
        if not pdb_proxy:
            pdb_proxy = PdbProxy(
                trace=self,
                thread_asynctask_id=thread_asynctask_id,
                modules_to_trace=self.modules_to_trace,
                registry=self.registry,
                ci_registry=self.pdb_ci_registry,
                prompting_counter=self.prompting_counter,
            )
            self.pdb_proxies[thread_asynctask_id] = pdb_proxy

        return pdb_proxy.trace_func(frame, event, arg)

    def returning(self, thread_asynctask_id):
        self.id_composer.exited(thread_asynctask_id)
        del self.pdb_proxies[thread_asynctask_id]


##__________________________________________________________________||



"""
Test against the issue
https://github.com/simonsobs/nextline/issues/7
"""


import asyncio
from pathlib import Path

import pytest

from nextline import Nextline

##__________________________________________________________________||
statement = """
import time
def f():
    time.sleep(1.0)
f()
""".strip()


##__________________________________________________________________||
@pytest.fixture(autouse=True)
def monkey_patch_syspath(monkeypatch):
    this_dir = Path(__file__).resolve().parent
    monkeypatch.syspath_prepend(str(this_dir))
    yield


##__________________________________________________________________||
async def monitor_global_state(nextline):
    async for s in nextline.subscribe_global_state():
        print("monitor_global_state()", s)


async def control_execution(nextline):
    prev_ids = set()
    tasks = set()

    # The lines of code between ============= can be rewritten with
    # one line of code as
    #   async for ids in nextline.subscribe_thread_asynctask_ids():
    # if exceptions occurred in tasks don't need to be re-raised.

    # ==================================================
    subscription = nextline.subscribe_thread_asynctask_ids()
    task_anext = asyncio.create_task(subscription.__anext__())
    while True:
        aws = {task_anext, *tasks}
        done, pending = await asyncio.wait(
            aws, return_when=asyncio.FIRST_COMPLETED
        )
        results = [  # noqa: F841
            t.result() for t in tasks if t in done
        ]  # re-raise exception
        tasks = tasks & pending
        if not task_anext.done():
            continue
        try:
            ids = task_anext.result()
        except StopAsyncIteration:
            break
        task_anext = asyncio.create_task(subscription.__anext__())
        # ===============================================

        ids = set(ids)

        new_ids = ids - prev_ids

        for id_ in new_ids:
            task = asyncio.create_task(control_thread_task(nextline, id_))
            tasks.add(task)

        prev_ids = ids


async def control_thread_task(nextline, thread_task_id):
    to_step = []
    print(f"control_thread_task({thread_task_id})")
    file_name = ""
    async for s in nextline.subscribe_thread_asynctask_state(thread_task_id):
        # print(s)
        if not file_name == s["file_name"]:
            file_name = s["file_name"]
            assert nextline.get_source(file_name)
        if s["prompting"]:
            command = "next"
            if s["trace_event"] == "line":
                line = nextline.get_source_line(
                    line_no=s["line_no"], file_name=s["file_name"]
                )
                if line in to_step:
                    # print(line)
                    command = "step"
            nextline.send_pdb_command(thread_task_id, command)


async def run(nextline):
    await asyncio.sleep(0.1)
    nextline.run()
    await nextline.finish()
    nextline.exception()
    nextline.result()  # raise exception
    await nextline.close()


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_run():

    nextline = Nextline(statement)

    assert nextline.global_state == "initialized"

    task_monitor_global_state = asyncio.create_task(
        monitor_global_state(nextline)
    )

    task_control_execution = asyncio.create_task(control_execution(nextline))

    task_run = asyncio.create_task(run(nextline))

    aws = [task_run, task_monitor_global_state, task_control_execution]
    await asyncio.gather(*aws)

    assert nextline.global_state == "closed"


##__________________________________________________________________||



import linecache

from .state import Machine


# __________________________________________________________________||
class Nextline:
    """Nextline allows line-by-line execution of concurrent Python scripts

    Nextline supports concurrency with threading and asyncio. It uses
    multiple instances of Pdb, one for each thread and async task.

    Note
    ----
    The running asyncio event loop must exists when Nextline is
    instantiated.

    Parameters
    ----------
    statement : str
        A Python code as a string

    """

    def __init__(self, statement):
        self.machine = Machine(statement)
        self.registry = self.machine.registry

    def __repr__(self):
        # e.g., "<Nextline 'running'>"
        return f"<{self.__class__.__name__} {self.global_state!r}>"

    @property
    def global_state(self) -> str:
        """state, e.g., "initialized", "running", "exited", "finished", "closed" """
        return self.machine.state_name

    def run(self):
        """execute the script"""
        self.machine.run()

    async def finish(self):
        """wait until the script execution exits

        wait for the script execution in another thread to exit and
        join the thread.

        """
        await self.machine.finish()

    def exception(self):
        """uncaught exeption from the last run"""
        return self.machine.exception()

    def result(self):
        """return value of the last run. always None"""
        self.machine.result()

    def reset(self, statement=None):
        """prepare for run"""
        self.machine.reset(statement=statement)

    async def close(self):
        """end gracefully"""
        await self.machine.close()

    async def subscribe_global_state(self):
        # wish to be able to write with "yield from" but not possible
        # https://stackoverflow.com/a/59079548/7309855
        agen = self.registry.subscribe("state_name")
        async for y in agen:
            yield y

    async def subscribe_thread_asynctask_ids(self):
        agen = self.registry.subscribe("thread_task_ids")
        async for y in agen:
            yield y

    async def subscribe_thread_asynctask_state(self, thread_asynctask_id):
        agen = self.registry.subscribe(thread_asynctask_id)
        async for y in agen:
            yield y

    def send_pdb_command(self, thread_asynctask_id, command):
        self.machine.send_pdb_command(thread_asynctask_id, command)

    def get_source(self, file_name=None):
        if not file_name or file_name == self.registry.get("script_file_name"):
            return self.registry.get("statement").split("\n")
        return [e.rstrip() for e in linecache.getlines(file_name)]

    def get_source_line(self, line_no, file_name=None):
        """
        based on linecache.getline()
        https://github.com/python/cpython/blob/v3.9.5/Lib/linecache.py#L26
        """
        lines = self.get_source(file_name)
        if 1 <= line_no <= len(lines):
            return lines[line_no - 1]
        return ""


# __________________________________________________________________||



from functools import partial
from pdb import Pdb


##__________________________________________________________________||
def getlines(func_org, statement, filename, module_globals=None):
    if filename == "<string>":
        return statement.split("\n")
    return func_org(filename, module_globals)


##__________________________________________________________________||
class CustomizedPdb(Pdb):
    """A customized Pdb

    An instance of this class will be created for each thread and async task

    """

    def __init__(self, proxy, prompting_counter, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._proxy = proxy
        self._prompting_counter = prompting_counter

        # self.quitting = True # not sure if necessary

        # stop at the first line
        self.botframe = None
        self._set_stopinfo(None, None)

    def trace_dispatch(self, frame, event, arg):
        """override trace_dispatch()"""
        self._trace_event = event
        # print(event)
        # if event == 'exception':
        #     print(arg)
        #     print(type(arg))
        #     print([type(e) for e in arg])
        return super().trace_dispatch(frame, event, arg)

    def _cmdloop(self):
        frame = self.curframe
        # module_name = frame.f_globals.get("__name__")
        state = {
            "prompting": self._prompting_counter(),
            "file_name": self.canonic(frame.f_code.co_filename),
            "line_no": frame.f_lineno,
            "trace_event": self._trace_event,
        }
        self._proxy.entering_cmdloop(self.curframe, state)
        super()._cmdloop()
        state["prompting"] = 0
        self._proxy.exited_cmdloop(state)

    def set_continue(self):
        """override bdb.set_continue()

        To avoid sys.settrace(None) called in bdb.set_continue()

        """
        self._set_stopinfo(self.botframe, None, -1)

    # def break_here(self, frame):
    #     ret = super().break_here(frame)
    #     print('break_here', ret)
    #     return ret

    # def stop_here(self, frame):
    #     ret = super().stop_here(frame)
    #     print('stop_here', ret)
    #     return ret

    # def bp_commands(self, frame):
    #     return True

    def do_list(self, arg):
        statement = self._proxy.statement
        import linecache

        getlines_org = linecache.getlines
        linecache.getlines = partial(getlines, getlines_org, statement)
        try:
            return super().do_list(arg)
        finally:
            linecache.getlines = getlines_org


##__________________________________________________________________||



import sys
import asyncio

import pytest
from unittest.mock import AsyncMock

from nextline.utils import CoroutineRunner


##__________________________________________________________________||
@pytest.fixture()
async def obj():
    y = CoroutineRunner()
    yield y


@pytest.mark.asyncio
async def test_repr(obj):
    repr(obj)


@pytest.mark.asyncio
async def test_async(obj):
    afunc = AsyncMock()
    coro = afunc()
    task = obj.run(coro)
    await task
    assert 1 == afunc.await_count


@pytest.mark.skipif(sys.version_info < (3, 9), reason="asyncio.to_thread()")
@pytest.mark.asyncio
async def test_thread(obj):
    def func():
        afunc = AsyncMock()
        coro = afunc()
        ret = obj.run(coro)
        assert ret is None
        assert 1 == afunc.await_count

    await asyncio.to_thread(func)


##__________________________________________________________________||
@pytest.mark.skipif(sys.version_info < (3, 9), reason="asyncio.to_thread()")
@pytest.mark.asyncio
async def test_error_no_loop():
    def func():
        # in a thread without an event loop
        with pytest.raises(RuntimeError):
            _ = CoroutineRunner()

    await asyncio.to_thread(func)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="asyncio.to_thread()")
@pytest.mark.asyncio
async def test_error_loop_closed():
    async def instantiate():
        obj = CoroutineRunner()
        return obj

    def func():

        # initialize the class while the event loop is running
        obj = asyncio.run(instantiate())

        # the loop is closed

        afunc = AsyncMock()
        coro = afunc()
        with pytest.raises(RuntimeError):
            obj.run(coro)

        assert 0 == afunc.await_count

    with pytest.warns(RuntimeWarning):
        # RuntimeWarning: coroutine 'AsyncMockMixin._execute_mock_call' was never awaited
        await asyncio.to_thread(func)



from .queuedist import QueueDist  # noqa: F401
from .thread_task_id import UniqThreadTaskIdComposer  # noqa: F401
from .thread_safe_event import ThreadSafeAsyncioEvent  # noqa: F401
from .coro_runner import CoroutineRunner  # noqa: F401
from .registry import Registry  # noqa: F401
